package it.polito.bigdata.spark.exercise2;
import org.apache.spark.api.java.*;

import scala.Tuple2;
import java.util.Map;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.spark.SparkConf;

public class SparkDriver {

    public static void main(String[] args) {
        
		Logger.getLogger("org").setLevel(Level.OFF);
		Logger.getLogger("akka").setLevel(Level.OFF);

        String purchasePath = "data/Purchases.txt";
        //String usersPath = "data/Users.txt";
        String cataloguePath = "data/Catalogue.txt";
        String outputPath1;
        String outputPath2;

        
        outputPath1 = "outPart1/";
        outputPath2 = "outPart2/";

        // Initialize SparkContext
        SparkConf conf = new SparkConf().setAppName("Exam 20240219_Spark").setMaster("local");
		// Create a Spark Context object
		JavaSparkContext sc = new JavaSparkContext(conf);


        // load all rdds
        // SaleTimestamp,UserID,ItemID,SalePrice
        JavaRDD<String> purchaseRDD = sc.textFile(purchasePath);
        // UserID,Name,Surname,City,Country
        //JavaRDD<String> usersRDD = sc.textFile(usersPath);
        // ItemID,Name,Category,StillInProduction
        JavaRDD<String> catalogueRDD = sc.textFile(cataloguePath);

        // ########################################
        // PART 1
        // ########################################
        // filter and keep only the purchases in year 2022 and year 2023
        JavaRDD<String> purchases2223 = purchaseRDD
                                        .filter(s -> s.startsWith("2023") || s.startsWith("2022"))
                                        .cache();

        // count for each user and year the number of purchases the user made
        // key = userId, year
        // value = +1
        // and use a reduceByKey to count the number of purchases
        JavaPairRDD<Tuple2<String, String>, Integer> userPurchasePerYear = 
                    purchases2223.mapToPair(s -> {
                        String fields[] = s.split(",");
                        String userId = fields[1];
                        String year = fields[0].split("/")[0];
                        Tuple2<String, String> key = new Tuple2<>(userId, year);
                        return new Tuple2<>(key, 1);
                    })
                    .reduceByKey((v1, v2) -> v1 + v2)
                    .cache();

        // compute for each year, the maximum number of purchases among all users
        // collect the result as a map
        // NOTE: we can do this because, by construction, the resulting map has only 2 entries with a single integer value in the value part!
        Map<String, Integer> maxPurchasePerYear = userPurchasePerYear
                                                .mapToPair(t -> new Tuple2<>(t._1()._2(), t._2()))
                                                .reduceByKey((v1, v2) -> (v1 > v2) ? v1 : v2)
                                                .collectAsMap();

        // filter the entries in userPurchasePerYear RDD by keeping the 
        //lines associated with value part == the highest number of purchases per year
        JavaRDD<String> res1 = userPurchasePerYear
                                .filter(v -> v._2() == maxPurchasePerYear.get(v._1()._2()))
                                .map(t -> t._1()._1())
				.distinct();

        res1.saveAsTextFile(outputPath1);
        
        // ########################################
        // PART 2 v2
        // ########################################
        // considering the purchases in year 2022/2023 (purchases2223 RDD)
        // we use a mapToPair with
        // key = itemID
        // value = userID
        // and a distinct to obtain the distinct user-product purchases,
        // a perform a mapToPair + reduceByKey to count for each itemID,
        // the number of distinct users who bought that item
        // key = itemID
        // value = numberOfDistinctUsersPurchases
        JavaPairRDD<String, Integer> itemDistinctUsersPurchases = purchases2223
                            .mapToPair(s -> {
                                String fields[] = s.split(",");
                                String userId = fields[1];
                                String itemId = fields[2];
                                return new Tuple2<>(itemId, userId);
                            })
                            .distinct()
                            .mapToPair(t -> new Tuple2<>(t._1(), 1))
                            .reduceByKey((v1, v2) -> v1 + v2);

        // for each item, we retrieve the corresponding category
        JavaPairRDD<String, String> itemCategory = catalogueRDD
                            .mapToPair(s -> {
                                String[] fields = s.split(",");
                                String itemId = fields[0];
                                String category = fields[2];
                                return new Tuple2<>(itemId, category);
                            });

        // we join itemCategory RDD with itemDistinctUsersPurchases
        // a leftOuterJoin is used to consider the 0-case,
        // i.e. the case in which ALL items in a category were not sold
        // consequently, we replace missing count values with 0
        // key = itemId
        // value = category, number of distinct users who purchased the item
        JavaPairRDD<String, Tuple2<String, Integer>> itemCategoryPurchases = itemCategory
                            .leftOuterJoin(itemDistinctUsersPurchases)
                            .mapToPair(t -> new Tuple2<>(t._1(), new Tuple2<>(t._2()._1(), t._2()._2().orElse(0))))
                            .cache();

        // compute for each category the maximum number distinct users who purchased the item
        // first, we obtain the following pairRDD
        // key = category
        // value = number of distinct users who purchased an item
        // and then we use a reduceByKey to compute the maximum value for each category
        JavaPairRDD<String, Integer> maxDistinctUsersPurchasesPerCategory = itemCategoryPurchases
                            .mapToPair(t -> new Tuple2<>(t._2()._1(), t._2()._2()))
                            .reduceByKey((v1, v2) -> v1 > v2 ? v1 : v2);

        // join itemCategoryPurchases with maxDistinctUsersPurchasesPerCategory
        // and filter, keeping only the entries with count == maxCount
        // after join, format is
        // key = category
        // value = (itemId, count), maxCountPerCategory
        // then, use a mapToPair to obtain the format for the result
        // key = category
        // value = itemId OR "NoPurchases"
        // final distinct is used to duplicated "NoPurchases"
        JavaPairRDD<String, String> res2 = itemCategoryPurchases
                            .mapToPair(t -> new Tuple2<>(t._2()._1(), new Tuple2<>(t._1(), t._2()._2())))
                            .join(maxDistinctUsersPurchasesPerCategory)
                            .filter(t -> t._2()._1()._2() == t._2()._2())
                            .mapToPair(t -> {
                                String category = t._1();
                                String value = t._2()._2() == 0 ? "NoPurchases" : t._2()._1()._1();
                                return new Tuple2<>(category, value);
                            })
                            .distinct();

        res2.saveAsTextFile(outputPath2);

        // Stop SparkContext
        sc.close();
    }
}
